{**************************************************************************************************}
{                                                                                                  }
{ AdsExtra, Version 1.1   Additional unit for ADS-Dataset                                          }
{                                                                                                  }
{  The contents of this file are subject to the Y Library Public License Version 1.0 (the          }
{  "License"); you may not use this file except in compliance with the License. You may obtain a   }
{  copy of the License at http://delphi.pjh2.de/                                                   }
{                                                                                                  }
{  Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF  }
{  ANY KIND, either express or implied. See the License for the specific language governing        }
{  rights and limitations under the License.                                                       }
{                                                                                                  }
{  The Original Code is: AdsExtra.pas.                                                             }
{  The Initial Developer of the Original Code is Peter J. Haas (libs@pjh2.de). Portions created    }
{  by Peter J. Haas are Copyright (C) 2000-2005 Peter J. Haas. All Rights Reserved.                }
{                                                                                                  }
{  Contributor(s):                                                                                 }
{                                                                                                  }
{  You may retrieve the latest version of this file at the homepage of Peter J. Haas, located at   }
{  http://delphi.pjh2.de/                                                                          }
{                                                                                                  }
{**************************************************************************************************}

// For history see end of file

{$ALIGN ON, $BOOLEVAL OFF, $LONGSTRINGS ON, $IOCHECKS ON, $WRITEABLECONST OFF}
{$OVERFLOWCHECKS OFF, $RANGECHECKS OFF, $TYPEDADDRESS ON, $MINENUMSIZE 1}
unit AdsExtra;

{$INCLUDE versions.inc}

interface
uses
  Windows, SysUtils, ACE;

{ Get ACE date format and convert to Delphi format }
function AdsDelphiDateFormat: String;

{ Set ACE date format to Delphi format }
procedure AdsSetDelphiDateFormat;

{ Convert ShortDateFormat to ACE date format }
function AdsGetLocalAdsDateFormat: String;

{ Convert TDateTime to string in ACE format }
function AdsDateToStr(AValue: TDateTime): String;
function AdsTimeToStr(AValue: TDateTime): String;
function AdsDateTimeToStr(AValue: TDateTime): String;

{ Convert TDateTime to string in SQL format }
function AdsSQLDateToStr(AValue: TDateTime): String;
function AdsSQLTimeToStr(AValue: TDateTime): String;
function AdsSQLDateTimeToStr(AValue: TDateTime): String;

{ Convert a SQL date string (CCYY-MM-DD and CCYYMMDD) to TDateTime }
function AdsSQLStrToDate(const AValue: String): TDateTime;

{ Convert a SQL time string (hh:mm:ss.z, hhmmss.z, hh:mm and hhmm) to TDateTime }
function AdsSQLStrToTime(const AValue: String): TDateTime;

{ Convert a SQL time stamp to a TDateTime,
  date and time are separated with a space }
function AdsSQLStrToDateTime(const AValue: String): TDateTime;

{ Convert float to string with . as decimal separator }
function AdsFloatToStr(AValue: Extended): String;
{ Convert string with . as decimal separator to float }
function AdsStrToFloat(const AValue: String): Extended;

{ Get server time stamp }
function AdsGetServerTimeStamp(AConnectionHandle: ADSHANDLE): TDateTime;

const
  AdsDelphiTimeFormat = 'hh":"nn":"ss AM/PM';
  AdsSQLDelphiDateFormat = 'YYYY"-"MM"-"DD';
  AdsSQLDelphiTimeFormat = 'hh":"nn":"ss"."z';
  AdsSQLDelphiDateTimeFormat = 'YYYY"-"MM"-"DD hh":"nn":"ss"."z';

const
  ChronicalJulianDayDelta = DateDelta + 1721425;

function DateToChronicalJulianDay(ADate: TDateTime): DWord;
function ChronicalJulianDayToDate(ADate: DWord): TDateTime;
function TimeToMilliseconds(ATime: TDateTime): DWord;
function MillisecondsToTime(ATime: DWord): TDateTime;

procedure ACECheck(AErrorCode: DWord);

type
  EACEError = class(Exception)
  private
    FErrorCode: Integer;
  public
    constructor Create(AErrorCode: Integer);
    property ErrorCode: Integer read FErrorCode;
  end;

implementation

{$ifndef ADSDELPHI3_OR_NEWER}
const
{$else}
resourcestring
{$endif}
  SInvalidSQLDate = '''%s'' is not a valid SQL date';
  SInvalidSQLTime = '''%s'' is not a valid SQL time';
  SInvalidSQLDateTime = '''%s'' is not a valid SQL date and time';
  SInvalidFloat = '''%s'' is not a valid floating point value';

  SACEError = 'ACE Error: %d'#$0D#$0A'%s';
  SACEUnknownError = 'Unknown error';

constructor EACEError.Create(AErrorCode: Integer);
var
  Buffer : array[0..ADS_MAX_ERROR_LEN] of Char;
  Len    : Word;
  Res    : Integer;
begin
  Len := SizeOf(Buffer);
  Res := ACE.AdsGetErrorString(AErrorCode, @Buffer, @Len);
  if Res <> AE_SUCCESS then
    StrPLCopy(Buffer, SACEUnknownError, SizeOf(Buffer));
  inherited CreateFmt(SACEError, [AErrorCode, Buffer]);
  FErrorCode := AErrorCode;
end;

procedure ACECheck(AErrorCode: DWord); register;

  function ReturnAddr: Pointer;
  asm
    {$ifopt W+}
          MOV     EAX, [EBP + 4]
    {$else}
          MOV     EAX, [ESP + 8]
    {$endif}
  end;

begin
  if AErrorCode <> AE_SUCCESS then
    raise EACEError.Create(AErrorCode) at ReturnAddr;
end;

function AdsDelphiDateFormat: String;
var
  s         : String;
  i         : Integer;
  Len       : Word;
  QuoteMode : Boolean;
begin
  { get DateFormat }
  Len := 0;
  AdsGetDateFormat(Nil, @Len);
  SetLength(s, Len-1);
  AdsGetDateFormat(Pointer(s), @Len);
  { convert DateFormat }
  Result := '';
  QuoteMode := False;
  s := UpperCase(s);
  for i := 1 to Length(s) do begin
    if s[i] in ['C'] then
      s[i] := 'Y';
    if (Not QuoteMode) xor (s[i] in ['D', 'M', 'Y']) then begin
      Result := Result + '"';
      QuoteMode := Not QuoteMode;
    end;
    Result := Result + s[i];
  end;
  if QuoteMode then
    Result := Result + '"';
end;

procedure AdsSetDelphiDateFormat;
begin
  ACECheck(ACE.AdsSetDateFormat(PChar(AdsGetLocalAdsDateFormat)));
end;


type
  TDateOrder = (doMDY, doDMY, doYMD);

function AdsGetLocalAdsDateFormat: String;
const
  Formats : array[TDateOrder] of String =
    ('MM%sDD%0:s%sYY', 'DD%sMM%0:s%sYY', '%1:sYY%0:sMM%0:sDD');
var
  s : String;
  i, Len, YearDigits : Integer;
  DateOrder : TDateOrder;
  DateSeparator : String;
  Century : String;
begin
  Result := '';
  s := ShortDateFormat;
  s := UpperCase(s);
  Len := Length(s);
  // determine order
  DateOrder := doMDY;
  i := 1;
  while i <= Len do begin
    case s[i] of
      'M' : ;
      'D' : DateOrder := doDMY;
      'Y' : DateOrder := doYMD;
    else
      Inc(i);
      Continue;
    end;
    Break;
  end;
  if i > Len then Exit;

  // determine year digit count
  i := Pos('Y',s);
  if i = 0 then Exit;
  YearDigits := 1;
  Inc(i);
  while (i <= Len) and (s[i]='Y') do begin
    Inc(i);
    Inc(YearDigits);
  end;
  if Not (YearDigits in [2, 4]) then Exit;

  // determine DateSeparator
  DateSeparator := '';
  i := 1;
  while i <= Len do begin
    if Not (s[i] in ['Y', 'M', 'D']) then begin
      DateSeparator := s[i];
      Break;
    end;
    Inc(i);
  end;

  // compound
  if YearDigits = 2 then
    Century := ''
  else
    Century := 'CC';
  Result := Format(Formats[DateOrder], [DateSeparator, Century]);
end;


function AdsDateToStr(AValue: TDateTime): String;
begin
  Result := FormatDateTime(AdsDelphiDateFormat, AValue);
end;

function AdsTimeToStr(AValue: TDateTime): String;
begin
  Result := FormatDateTime(AdsDelphiTimeFormat, AValue);
end;

function AdsDateTimeToStr(AValue: TDateTime): String;
begin
  Result := FormatDateTime(AdsDelphiDateFormat + ' ' + AdsDelphiTimeFormat, AValue);
end;

function AdsSQLDateToStr(AValue: TDateTime): String;
begin
  Result := FormatDateTime(AdsSQLDelphiDateFormat, AValue);
end;

{$ifdef ADSDELPHI5_OR_NEWER}  { poss. since delphi 4 }
function AdsSQLTimeToStr(AValue: TDateTime): String;
begin
  Result := FormatDateTime(AdsSQLDelphiTimeFormat, AValue);
end;
{$else}
function AdsSQLTimeToStr(AValue: TDateTime): String;
var
  Hours, Minutes, Seconds, Milliseconds : Word;
begin
  DecodeTime(AValue, Hours, Minutes, Seconds, Milliseconds);
  Result:=Format('%.2d:%.2d:%.2d.%d', [Hours, Minutes, Seconds, Milliseconds]);
end;
{$endif}

{$ifdef ADSDELPHI5_OR_NEWER}  { poss. since delphi 4 }
function AdsSQLDateTimeToStr(AValue: TDateTime): String;
begin
  Result := FormatDateTime(AdsSQLDelphiDateTimeFormat, AValue);
end;
{$else}
function AdsSQLDateTimeToStr(AValue: TDateTime): String;
begin
  Result := AdsSQLDateToStr(AValue) + ' ' + AdsSQLTimeToStr(AValue);
end;
{$endif}

function CheckDateTimeFormat(const AValue, AFormat: String): Boolean;
var
  i : Integer;
begin
  Result := False;
  if Length(AValue) <> Length(AFormat) then Exit;
  for i:= 1 to Length(AValue) do begin
    if AFormat[i] = '9' then begin   { Digit }
      if Not (AValue[i] in ['0'..'9']) then Exit;
    end
    else begin
      if AFormat[i] <> AValue[i] then Exit;
    end;
  end;
  Result:= True;
end;

function ScanSQLDate(const AValue: String; var ADate: TDateTime): Boolean;
var
  Offset : Integer;
  Year, Month, Day : Word;
begin
  Result := False;
  if CheckDateTimeFormat(AValue, '9999-99-99') then begin
    Offset := 1;
  end
  else if CheckDateTimeFormat(AValue, '99999999') then begin
    Offset := 0;
  end
  else begin
    Exit;
  end;
  Year  := StrToIntDef(Copy(AValue, 1, 4), 0);
  Month := StrToIntDef(Copy(AValue, 5 + 1*Offset, 2), 0);
  Day   := StrToIntDef(Copy(AValue, 7 + 2*Offset, 2), 0);
  try
    ADate := EncodeDate(Year, Month, Day) + Frac(ADate);
    Result := True;
  except
    on EConvertError do;
  end;
end;


function ScanSQLTime(const AValue: String; var ADate: TDateTime): Boolean;
var
  s, ms : String;
  i : Integer;
  Hours, Minutes, Seconds, Milliseconds : Word;
  Offset : Integer;
  WithSeconds : Boolean;

begin
  Result := False;
  i := Pos('.', AValue);
  if i = 0 then begin
    s := AValue;
    ms := '';
  end
  else begin
    s := Copy(AValue, 1, i-1);
    ms := Copy(AValue, i+1, MaxInt);
  end;
  if CheckDateTimeFormat(s, '99:99:99') then begin
    Offset := 1;
    WithSeconds := True;
  end
  else if CheckDateTimeFormat(s, '999999') then begin
    Offset := 0;
    WithSeconds := True;
  end
  else if CheckDateTimeFormat(s, '99:99') then begin
    Offset := 1;
    WithSeconds := False;
  end
  else if CheckDateTimeFormat(s, '9999') then begin
    Offset := 0;
    WithSeconds := False;
  end
  else begin
    Exit;
  end;

  Hours   := StrToIntDef(Copy(AValue, 1, 2), 100);
  Minutes := StrToIntDef(Copy(AValue, 3 + 1*Offset, 2), 100);

  if WithSeconds then begin
    Seconds := StrToIntDef(Copy(AValue, 5 + 2*Offset, 2), 100);
  end
  else begin
    Seconds := 0;
    if Length(ms)>0 then Exit;  { Milliseconds without seconds -> error }
  end;

  if Length(ms)>0 then begin
    Milliseconds := StrToIntDef(ms, 10000);
  end
  else begin
    Milliseconds := 0;
  end;

  try
    ADate := EncodeTime(Hours, Minutes, Seconds, Milliseconds) + Int(ADate);
    Result := True;
  except
    on EConvertError do;
  end;
end;

function AdsSQLStrToDate(const AValue: String): TDateTime;
begin
  Result := 0;
  if Not ScanSQLDate(AValue, Result) then
    raise EConvertError.CreateFmt(SInvalidSQLDate, [AValue]);
end;

function AdsSQLStrToTime(const AValue: String): TDateTime;
begin
  Result := 0;
  if Not ScanSQLTime(AValue, Result) then
    raise EConvertError.CreateFmt(SInvalidSQLTime, [AValue]);
end;

function AdsSQLStrToDateTime(const AValue: String): TDateTime;
var
  ADate, ATime : String;
  i : Integer;
begin
  Result := 0;
  i := Pos(' ', AValue);
  if i > 0 then begin
    ADate := Copy(AValue, 1, i-1);
    ATime := Copy(AValue, i+1, MaxInt);
    if ScanSQLDate(ADate, Result) and ScanSQLTime(ATime, Result) then Exit;
  end;
  raise EConvertError.CreateFmt(SInvalidSQLDateTime, [AValue]);
end;

function AdsFloatToStr(AValue: Extended): String;
var
  Decimals : Word;
begin
  ACE.AdsGetDecimals(@Decimals);
  Str(AValue:Decimals:Decimals, Result);
end;

function AdsStrToFloat(const AValue: String): Extended;
var
  Error : Integer;
begin
  Val(AValue, Result, Error);  { DecimalSeparator always '.' }
  if Error <> 0 then
    raise EConvertError.CreateFmt(SInvalidFloat, [AValue]);
end;


function DateToChronicalJulianDay(ADate: TDateTime): DWord;
begin
  Result := Trunc(ADate) + ChronicalJulianDayDelta;
end;

function ChronicalJulianDayToDate(ADate: DWord): TDateTime;
begin
  Result := ADate - ChronicalJulianDayDelta;
end;

function TimeToMilliseconds(ATime: TDateTime): DWord;
begin
  Result := Trunc(Frac(ATime) * MSecsPerDay);
end;

function MillisecondsToTime(ATime: DWord): TDateTime;
begin
  Result := ATime / MSecsPerDay;
end;

function AdsGetServerTimeStamp(AConnectionHandle: ADSHANDLE): TDateTime;
var
  DateFormatBuffer    : array[0..12] of Char;
  DateFormatBufferLen : Word;
  DateBuffer    : array[0..12] of Char;
  DateBufferLen : Word;
  Time          : LongInt;
  TimeBuffer    : array[0..12] of Char;
  TimeBufferLen : Word;
begin
  // Save ACE date format
  DateFormatBufferLen := SizeOf(DateFormatBuffer);
  ACECheck(ACE.AdsGetDateFormat(@DateFormatBuffer, @DateFormatBufferLen));
  try
    // Set ACE date format to Delphi date format
    ACECheck(ACE.AdsSetDateFormat(PChar(AdsGetLocalAdsDateFormat)));
    // Get server time stamp
    DateBufferLen := SizeOf(DateBuffer);
    TimeBufferLen := SizeOf(TimeBuffer);
    ACECheck(ACE.AdsGetServerTime(AConnectionHandle,
                                  @DateBuffer, @DateBufferLen,
                                  @Time,
                                  @TimeBuffer, @TimeBufferLen));
    // Convert to TDateTime
    Result := StrToDate(DateBuffer) + AdsExtra.MillisecondsToTime(Time);
  finally
    // Restore ACE date format
    ACECheck(ACE.AdsSetDateFormat(@DateFormatBuffer));
  end;
end;

// *******************************************************************************************

//  History:
//   2005-02-20, Peter J. Haas
//    - new license
//
//   2001-01-17, Peter J. Haas, Version 1.2
//    - add AdsSetDelphiDateFormat, AdsGetLocalAdsDateFormat
//    - add EACEError, ACECheck
//
//   2000-12-31, Peter J. Haas, Version 1.1
//    - add DateToChronicalJulianDay, ChronicalJulianDayToDate
//    - add TimeToMilliseconds, MillisecondsToTime
//
//   2000-11-23, Peter J. Haas, Version 1.0
//    - First public version

end.
